﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Security.Cryptography;
using System.Text.RegularExpressions;
using System.Xml.Linq;

namespace CommonLib
{
    /// <summary>
    /// 用来写常用的扩展方法
    /// </summary>
    public static class OperationObject
    {
        public static string GetAttributeValue(this XElement xml, string code)
        {
            if (xml == null) return "";

            var node = xml.Attribute(code);

            return node == null ? "" : node.Value ?? "";
        }

        public static string GetElementValue(this XElement xml, string code)
        {
            if (xml == null) return "";

            var node = xml.Element(code);

            return node == null ? "" : node.Value ?? "";
        }

        /// <summary>
        /// 获取分页数据
        /// </summary>
        /// <typeparam name="T">当前数据类型</typeparam>
        /// <param name="source">源数据</param>
        /// <param name="pageIndex">要返回的结果页的索引。使用 0 来指示第一页,pageIndex 是从零开始的。</param>
        /// <param name="pageSize">要返回的结果页的大小。</param>
        /// <param name="total">记录的总数</param>
        /// <returns></returns>
        /// <exception cref="ArgumentException">source不可为空</exception>
        /// <exception cref="ArgumentException">pageIndex 小于零。- 或 -pageSize 小于 1。</exception>
        public static IQueryable<T> GetPageList<T>(this IQueryable<T> source, int pageIndex, int pageSize, out int total)
        {
            if (source == null) throw new ArgumentNullException("source不可为空");

            if (pageIndex < 0 || pageSize < 1) throw new ArgumentException("pageIndex 小于零。- 或 -pageSize 小于 1");

            total = source.Count();

            return source.Skip(pageSize * pageIndex).Take(pageSize);
        }

        /// <summary>
        /// 获取分页数据
        /// </summary>
        /// <typeparam name="T">当前数据类型</typeparam>
        /// <param name="source">源数据</param>
        /// <param name="pageIndex">要返回的结果页的索引。使用 0 来指示第一页,pageIndex 是从零开始的。</param>
        /// <param name="pageSize">要返回的结果页的大小。</param>
        /// <param name="total">记录的总数</param>
        /// <returns></returns>
        /// <exception cref="ArgumentException">source不可为空</exception>
        /// <exception cref="ArgumentException">pageIndex 小于零。- 或 -pageSize 小于 1。</exception>
        public static List<T> GetPageList<T>(this IList<T> source, int pageIndex, int pageSize, out int total)
        {
            if (source == null) throw new ArgumentNullException("source不可为空");

            if (pageIndex < 0 || pageSize < 1) throw new ArgumentException("pageIndex 小于零。- 或 -pageSize 小于 1");

            total = source.Count();

            return source.Skip(pageSize * pageIndex).Take(pageSize).ToList();
        }

        /// <summary>
        /// 添加，如果Key已存在，则不添加，无异常
        /// </summary>
        /// <typeparam name="TKey"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="dic"></param>
        /// <param name="key"></param>
        /// <param name="value"></param>
        public static void AddData<TKey, TValue>(this Dictionary<TKey, TValue> dic, TKey key, TValue value)
        {
            if (dic == null) return;

            if (!dic.Keys.Contains(key))
            {
                dic.Add(key, value);
            }
        }

        /// <summary>
        /// 生成表达式，结构为 x=>x.id==xx||x.id==yy
        /// </summary>
        /// <typeparam name="TElement"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="source"></param>
        /// <param name="propertySelector"></param>
        /// <param name="values"></param>
        /// <returns></returns>
        public static IQueryable<TElement> WhereIn<TElement, TValue>(this IQueryable<TElement> source, Expression<Func<TElement, TValue>> propertySelector, params TValue[] values)
        {
            return source.Where(BuildWhereInExpression(propertySelector, values));
        }

        /// <summary>
        /// 生成sql语句中in的表达式，如where id in (2,3,4)
        /// </summary>
        /// <typeparam name="TElement"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="propertySelector">列名</param>
        /// <param name="values">集合</param>
        /// <returns></returns>
        private static Expression<Func<TElement, bool>> BuildWhereInExpression<TElement, TValue>(Expression<Func<TElement, TValue>> propertySelector, IEnumerable<TValue> values)
        {
            ParameterExpression p = propertySelector.Parameters.Single();
            if (!values.Any())
                return e => false;

            var equals = values.Select(value => (Expression)Expression.Equal(propertySelector.Body, Expression.Constant(value, typeof(TValue))));
            var body = equals.Aggregate<Expression>((accumulate, equal) => Expression.Or(accumulate, equal));

            return Expression.Lambda<Func<TElement, bool>>(body, p);
        }

        /// <summary>
        /// 转化成特定类型，该类型必须为结构类型，如果转化失败则返回T的默认值 
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="str"></param>
        /// <returns></returns>
        public static T ToStruct<T>(this string str) where T : struct
        {
            var defaultValue = default(T);

            if (string.IsNullOrEmpty(str)) return defaultValue;

            if (typeof(T).IsEnum)
            {
                Enum.TryParse(str, out defaultValue);

                return defaultValue;

            }
            else
            {
                var method = typeof(T).GetMethods().Where(x => x.Name == "TryParse" && x.GetParameters().Count() == 2).FirstOrDefault();

                if (method == null) return defaultValue;

                var objects = new object[] { str, defaultValue };

                method.Invoke(null, objects);

                return (T)objects[1];
            }
        }

        /// <summary>
        /// 时间格式
        /// </summary>
        public static string DateFormat = "yyyy-MM-dd";


        public static string DateFormatForBib = "yyyyMMdd";

        /// <summary>
        /// 格式化时间,格式为：2011-11-11
        /// </summary>
        /// <param name="time"></param>
        /// <returns></returns>
        public static string GetFormatForNullTime(this  DateTime? time)
        {
            return GetFormatForNullTime(time, DateFormat);
        }

        public static string GetFormatForNullTime(this DateTime? time, string timeFormat)
        {
            if (time == null || string.IsNullOrEmpty(timeFormat)) return string.Empty;
            return time.Value.ToString(timeFormat);

        }

        /// <summary>
        /// 格式化时间,格式为：2011-11-11
        /// </summary>
        /// <param name="time"></param>
        /// <returns></returns>
        public static string GetFormatTime(this DateTime time)
        {
            return GetFormatForNullTime(time);
        }
        /// <summary>
        /// 格式化时间,格式为：20111111
        /// </summary>
        /// <param name="time"></param>
        /// <returns></returns>
        public static string GetFormatTimeForBib(this DateTime time)
        {
            return GetFormatForNullTime(time, DateFormatForBib);
        }

        /// <summary>
        /// 创建字典
        /// </summary>
        /// <typeparam name="TKey"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="dic"></param>
        /// <param name="key"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static Dictionary<TKey, TValue> Create<TKey, TValue>(this Dictionary<TKey, TValue> dic, TKey key, TValue value)
        {
            dic.Add(key, value);
            return dic;

        }

        /// <summary>
        /// 添加，如果记录存在，则添加一个新的Key，否则追加到某一个Key中
        /// </summary>
        /// <typeparam name="TKey"></typeparam>
        /// <param name="dic"></param>
        /// <param name="key"></param>
        /// <param name="value"></param>
        public static void AddOrContainData<TKey>(this Dictionary<TKey, List<TKey>> dic, TKey key, TKey value)
        {
            if (!dic.Keys.Contains(key))
            {
                dic.Add(key, new List<TKey>() { value });
            }
            else
            {
                dic[key].Add(value);
            }
        }

        /// <summary>
        /// 根据字典的键来取值，如果值不存在，则返回默认值 
        /// </summary>
        /// <typeparam name="TKey"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="dic"></param>
        /// <param name="key"></param>
        /// <returns></returns>
        public static TValue GetValue<TKey, TValue>(this Dictionary<TKey, TValue> dic, TKey key)
        {
            if (dic.Keys.Contains(key)) return dic[key];

            return default(TValue);
        }

        public static string CustomReplace(this string value, string oldValue, string newValue)
        {
            if (string.IsNullOrEmpty(oldValue))
            {
                return value;
            }

            return value.Replace(oldValue, newValue);
        }




        /// <summary>
        /// 根据字典的值来取键，如果键不存在，则返回默认值 
        /// </summary>
        /// <typeparam name="TKey"></typeparam>
        /// <typeparam name="TValue"></typeparam>
        /// <param name="dic"></param>
        /// <param name="key"></param>
        /// <returns></returns>
        public static TKey GetKey<TKey, TValue>(this Dictionary<TKey, TValue> dic, TValue value)
        {
            return dic.Where(x => x.Value.Equals(value)).Select(x => x.Key).FirstOrDefault();
        }

        /// <summary>
        /// 把字符串转化整形,无异常，如果转化失败，则返回0
        /// </summary>
        /// <param name="str"></param>
        /// <returns></returns>
        public static int ToInt(this string str)
        {
            int i = 0;

            int.TryParse(str, out i);

            return i;
        }

        /// <summary>
        /// 把字符串转化十进制类型,无异常，如果转化失败，则返回0
        /// </summary>
        /// <param name="str"></param>
        /// <returns></returns>
        public static decimal ToDecimal(this string str)
        {
            decimal i = 0;

            decimal.TryParse(str, out i);

            return i;
        }

        /// <summary>
        /// 合并集合中的集合,把IEnumerable(IEnumerable(T))合并为一个IEnumerable(T)的集合
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <returns></returns>
        public static IEnumerable<T> Marge<T>(this IEnumerable<IEnumerable<T>> source)
        {
            IEnumerable<T> list = new List<T>();

            foreach (var item in source)
            {
                list = list.Concat(item);
            }

            return list;
        }

        /// <summary>
        /// 获取特定格式元素的值，如果元素长度不够，则低位补充elementChar
        /// </summary>
        /// <param name="value">元素值</param>
        /// <param name="len">长度</param>
        /// <param name="elementChar">要补充的值</param>
        /// <returns></returns>
        public static string GetFormatValue(this  string value, char elementChar, int len)
        {
            if (value.Length < len)
            {
                return value + new string(elementChar, len - value.Length);
            }
            return value;
        }

        /// <summary>
        /// 获取特定格式元素的值，如果元素长度不够，则高位补充elementChar
        /// </summary>
        /// <param name="value">元素值</param>
        /// <param name="len">长度</param>
        /// <param name="elementChar">要补充的值</param>
        /// <returns></returns>
        public static string GetFormatValueForUpper(this  string value, char elementChar, int len)
        {
            if (value.Length < len)
            {
                return new string(elementChar, len - value.Length) + value;
            }
            return value;
        }

        /// <summary>
        /// 获取指定格式的字符串，把集合中的对象，通过定界符来重新生成一个字符串。
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="collection"></param>
        /// <param name="delimiter">定界符</param>
        /// <returns></returns>
        public static string GetFormatString<T>(this ICollection<T> collection, string delimiter)
        {
            if (collection == null || collection.Count == 0) return string.Empty;

            return string.Join(delimiter, collection.Select(x => x.ToString()).ToArray());
        }

        /// <summary>
        /// 获取指定格式的字符串，把集合中的对象，通过定界符来重新生成一个字符串。
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="collection"></param>
        /// <param name="delimiter">定界符</param>
        /// <returns></returns>
        /// <exception cref="ArgumentException">当前集合中的字符串，包含了定界符</exception>
        public static string GetFormatString<T>(this ICollection<T> collection, char delimiter)
        {
            return collection.GetFormatString(delimiter.ToString());
        }

        /// <summary>
        /// 截取字符串，并追加相应的字符串
        /// </summary>
        /// <param name="str"></param>
        /// <param name="length">要截取的长度</param>
        /// <param name="appendString">追加的字符串</param>
        /// <param name="isDiscriminateCnOrEn">是否区分中英文</param>
        /// <returns></returns>
        public static string GetSubStringAndAppendString(this string str, int length, string appendString, bool isDiscriminateCnOrEn)
        {
            if (str == null || str.Length == 0) return appendString;

            if (isDiscriminateCnOrEn)
            {
                var encoding = Encoding.Default;

                var bytes = encoding.GetBytes(str);

                if (bytes.Length <= length)
                {
                    return str;
                }

                //如果为奇数
                if (length % 2 == 1)
                {
                    var temp = encoding.GetString(bytes, 0, length - 1);
                    return temp + str.Substring(temp.Length, 1) + appendString;
                }

                return encoding.GetString(bytes, 0, length) + appendString;
            }
            else
            {

                if (str.Length <= length) return str;

                return str.Substring(0, length) + appendString;

            }

        }

        /// <summary>
        /// 截取字符串，并追加相应的字符串,不区别中英文
        /// </summary>
        /// <param name="str"></param>
        /// <param name="length">要截取的长度</param>
        /// <param name="appendString">追加的字符串</param>
        /// <returns></returns>
        public static string GetSubStringAndAppendString(this string str, int length, string appendString)
        {
            return str.GetSubStringAndAppendString(length, appendString, false);

        }

        /// <summary>
        /// 把字符串变为md5密过的字符串
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public static String ToMd5(this String message)
        {
            Byte[] clearBytes = new UnicodeEncoding().GetBytes(message);
            Byte[] hashedBytes = ((HashAlgorithm)CryptoConfig.CreateFromName("MD5")).ComputeHash(clearBytes);
            String tt = BitConverter.ToString(hashedBytes).Replace("-", "");
            return tt;
        }

        /// <summary>
        /// 是否为数字
        /// </summary>
        /// <param name="stringValue"></param>
        /// <returns></returns>
        public static bool IsNumberic(this string stringValue)
        {
            Regex rex = new Regex(@"^\d+$");
            return rex.IsMatch(stringValue);
        }

        /// <summary>
        /// 是否为IP
        /// </summary>
        /// <param name="stringValue"></param>
        /// <returns></returns>
        public static bool IsIPAddress(this string stringValue)
        {
            Regex regex = new Regex(@"^(\d{1,2}|1\d\d|2[0-4]\d|25[0-5]).(\d{1,2}|1\d\d|2[0-4]\d|25[0-5]).(\d{1,2}|1\d\d|2[0-4]\d|25[0-5]).(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])$");
            return regex.IsMatch(stringValue);
        }
        /// <summary>
        /// 获取该对象的toString，如果对象为Null，则返回string.empty
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public static string GetObjectString(this  object obj)
        {
            if (obj == null) return string.Empty;

            return obj.ToString();
        }

        /// <summary>
        /// 截取指定长度，如果长度超过了字符串的最大长度，则从开始位截取到字符串的最后一位
        /// </summary>
        /// <param name="s"></param>
        /// <param name="start"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string CustomSubString(this string s, int start, int length)
        {
            if (string.IsNullOrEmpty(s)) return s;

            if (length <= 0) return s;

            return start + length > s.Length ? s.Substring(start) : s.Substring(start, length);
        }

        /// <summary>
        /// 截取指定长度，如果长度超过了字符串的最大长度，则从开始位截取到字符串的最后一位，区公中英文
        /// </summary>
        /// <param name="s"></param>
        /// <param name="start"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static string CustomSubStringNotCase(this string s, int start, int length)
        {
            if (string.IsNullOrEmpty(s)) return s;

            if (length <= 0) return s;

            var b = Encoding.UTF8.GetBytes(s);

            return start + length > b.Length ? s.Substring(start) : Encoding.UTF8.GetString(b, start, length - 1);
        }
    }


}
