﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.SqlClient;
using System.Web;

namespace System.IO.IOOperation
{
    /// <summary>
    /// 文件下载
    /// </summary>
    public interface IDownLoadFile
    {
        /// <summary>
        /// 每次读取文件的大小
        /// </summary>
        int BufferSize { get; set; }

        void DownLoad();
    }

    public class FileParms
    {
        public long Length { get; set; }

        public string FileName { get; set; }

        /// <summary>
        /// 文件扩展名，如:jpg,rar
        /// </summary>
        public string Format { get; set; }
    }

    /// <summary>
    /// 从数据库下载文件
    /// </summary>
    public class DownLoadFileForDataBase : IDownLoadFile
    {

        private string connection;

        private string select;

        private FileParms parms;

        /// <summary>
        /// 每次从数据库中读取流的大小，默认为:10kb 
        /// </summary>
        public int BufferSize
        {
            get;
            set;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="connection">数据库连接字符串</param>
        /// <param name="select">查询语句,只查询数据流，如：select contents from attachment where id=4</param>
        /// <param name="parms">文件参数</param>
        public DownLoadFileForDataBase(string connection, string select, FileParms parms)
        {
            if (string.IsNullOrEmpty(connection)) throw new ArgumentNullException("connection", "connection is null or empty");

            if (string.IsNullOrEmpty(select)) throw new ArgumentNullException("select", "select is null or empty");

            if (parms == null) throw new ArgumentNullException("parms", "parms is null");

            this.connection = connection;
            this.select = select;
            this.parms = parms;

            BufferSize = 10000;
        }

        #region DownLoadStream Members

        public void DownLoad()
        {
            SqlConnection conn = null;
            SqlCommand cmd = null;
            SqlDataReader reader = null;

            var context = HttpContext.Current;

            // context.Response.AddHeader("Cache-Control", "No-store");

            context.Response.AddHeader("Content-Disposition", "attachment; filename=\"" + HttpUtility.UrlEncode(parms.FileName + "." + parms.Format, Encoding.UTF8) + "\"");

            context.Response.AddHeader("Content-Length", parms.Length.ToString());

            context.Response.ContentType = FileMineType.FileExtensionToContentType.ContainsKey(parms.Format.ToLower()) ? FileMineType.FileExtensionToContentType[parms.Format.ToLower()] : FileMineType.DEFAULT_CONTENT_TYPE;

            try
            {
                conn = new SqlConnection(connection);

                cmd = new SqlCommand(select, conn);

                conn.Open();

                reader = cmd.ExecuteReader();

                var bufferSize = BufferSize;

                byte[] bytes = new byte[bufferSize];

                long startIndex = 0;

                long realSize = bufferSize;

                while (reader.Read())
                {
                    while (realSize == bufferSize)
                    {
                        if (context.Response.IsClientConnected)
                        {
                            realSize = reader.GetBytes(0, startIndex, bytes, 0, bufferSize);

                            context.Response.OutputStream.Write(bytes, 0, (int)realSize);

                            context.Response.Flush();

                            startIndex += bufferSize;

                            // Thread.Sleep(1000);
                        }
                        else break;
                    }
                    break;
                }
            }
            catch (Exception ex)
            {
                context.Response.Write("DownLoad is fail," + ex.Message);
            }
            finally
            {
                if (reader != null)
                    reader.Close();

                if (cmd != null)
                    cmd.Clone();

                if (conn.State == System.Data.ConnectionState.Open)
                    conn.Close();
            }
        }

        #endregion
    }

    public class DownLoadFileForDirectory : IDownLoadFile
    {

        private string filePath;

        public DownLoadFileForDirectory(string filePath)
        {
            if (string.IsNullOrEmpty(filePath) || !File.Exists(filePath)) throw new ArgumentNullException("filePath", "filePath is null or empty or not Exists");

            this.filePath = filePath;

            BufferSize = 10000;
        }

        public int BufferSize
        {
            get;
            set;
        }

        public void DownLoad()
        {
            System.IO.Stream iStream = null;

            var bufferSize = BufferSize;

            byte[] buffer = new Byte[bufferSize];

            int length;

            long dataToRead;

            string filename = System.IO.Path.GetFileName(filePath);

            string format = System.IO.Path.GetExtension(filePath).Replace(".", string.Empty);

            var response = HttpContext.Current.Response;

            try
            {
                iStream = new System.IO.FileStream(filePath, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.Read);

                dataToRead = iStream.Length;

                response.AddHeader("Content-Disposition", "attachment; filename=" + HttpUtility.UrlEncode(filename, System.Text.Encoding.UTF8));

                response.AddHeader("Content-Length", dataToRead.ToString());

                response.ContentType = FileMineType.FileExtensionToContentType.ContainsKey(format.ToLower()) ? FileMineType.FileExtensionToContentType[format.ToLower()] : FileMineType.DEFAULT_CONTENT_TYPE;

                while (dataToRead > 0)
                {
                    if (response.IsClientConnected)
                    {
                        length = iStream.Read(buffer, 0, bufferSize);

                        response.OutputStream.Write(buffer, 0, length);

                        response.Flush();

                        buffer = new Byte[bufferSize];

                        dataToRead = dataToRead - length;
                    }
                    else
                    {
                        dataToRead = -1;
                    }
                }
            }
            catch (Exception ex)
            {
                response.Write("文件下载时出现错误!" + ex.Message);
            }
            finally
            {
                if (iStream != null)
                {
                    iStream.Close();
                }
            }

        }
    }
}
