﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Linq.Expressions;
using System.Reflection;
using System.Collections.ObjectModel;
namespace System.Linq.LinqExtensions
{
    public static class ExpressionExpand
    {
        public static Expression Like(this Expression left, Expression right)
        {
            var value = ((System.Linq.Expressions.ConstantExpression)right).Value as string;

            right = RemovePrecent(value);

            if (value.StartsWith("%") && value.EndsWith("%"))
            {
                MethodInfo method = left.Type.GetMethod("Contains");
                return Expression.Call(left, method, new[] { right });
            }
            else if (value.StartsWith("%"))
            {
                MethodInfo method = left.Type.GetMethod("StartsWith", new[] { typeof(string) });
                return Expression.Call(left, method, new[] { right });
            }
            else if (value.EndsWith("%"))
            {
                MethodInfo method = left.Type.GetMethod("EndsWith", new[] { typeof(string) });
                return Expression.Call(left, method, new[] { right });
            }

            return Expression.Equal(left, right);
        }

        public static Expression NotLike(Expression left, Expression right)
        {
            return Expression.Not(Like(left, right));
        }

        public static Expression In(Expression left, Expression right)
        {
            return Expression.Call(right, right.Type.GetMethod("Contains"), new[] { left });
        }

        public static Expression NotIn(Expression left, Expression right)
        {
            return Expression.Not(In(left, right));
        }

        //去掉查询字符中的%符号
        private static Expression RemovePrecent(string value)
        {
            return Expression.Constant(value.Replace("%", String.Empty), typeof(string));
        }
    }

    public class SearchOperation
    {
        public string Column { get; set; }

        public Operator operatorType { get; set; }

        public object value { get; set; }

        public Logic logic { get; set; }

    }

    public enum Logic
    {
        and,
        or
    }

    public enum Operator
    {
        Equal,
        NotEqual,
        Like,
        NotLike,
        In,
        NotIn,
        GreaterThan,
        GreaterThanOrEqual,
        LessThan,
        LessThanOrEqual
    }

    public class OrderBy
    {
        public string Coloumn { get; set; }
        public bool IsDesc { get; set; }
    }

    public static class LambdaForSearch
    {
        public static IQueryable<T> Where<T>(this IQueryable<T> sources, ReadOnlyCollection<SearchOperation> pred) where T : class
        {
            if (sources == null) throw new ArgumentNullException();

            var express = BulidExpression<T>(pred);

            return sources.Where(express);
        }

        public static IQueryable<T> OrderBy<T>(this IQueryable<T> sources, ReadOnlyCollection<OrderBy> orderBy)
        {
            if (sources == null) throw new ArgumentNullException();

            if (orderBy == null || orderBy.Count == 0) return sources;

            ParameterExpression parm = Expression.Parameter(typeof(T), "");

            Expression queryExpr = sources.Expression;

            string methodAsc = "OrderBy";
            string methodDesc = "OrderByDescending";

            foreach (var item in orderBy)
            {
                var memberInfo = typeof(T).GetMember(item.Coloumn).FirstOrDefault();

                var ex = Expression.MakeMemberAccess(parm, memberInfo);

                queryExpr = Expression.Call(
                    typeof(Queryable), item.IsDesc ? methodDesc : methodAsc,
                    new Type[] { sources.ElementType, ex.Type },
                    queryExpr, Expression.Quote(Expression.Lambda(ex, parm)));

                methodAsc = "ThenBy";
                methodDesc = "ThenByDescending";

            }

            return (IQueryable<T>)sources.Provider.CreateQuery(queryExpr);
        }

        public static Expression<Func<T, bool>> BulidExpression<T>(ReadOnlyCollection<SearchOperation> pred)
        {
            Expression express = null;

            var parm = Expression.Parameter(typeof(T), "s");

            foreach (var p in pred)
            {
                Expression left = Expression.Property(parm, typeof(T).GetProperty(p.Column));

                var right = Expression.Constant(p.value, p.value.GetType());

                Expression filter = null;

                filter = ExpressFilter(p.operatorType, left, right);

                if (express == null)
                    express = filter;
                else
                {
                    if (p.logic == Logic.and)
                    {
                        express = Expression.And(express, filter);
                    }
                    else
                    {
                        express = Expression.Or(express, filter);
                    }
                }
            }

            return Expression.Lambda<Func<T, bool>>(express, parm);
        }

        public static Expression<Func<T, bool>> Or<T>(this Expression<Func<T, bool>> expression1, Expression<Func<T, bool>> expression2)
        {
            var invokedExpression = Expression.Invoke(expression2, expression1.Parameters.Cast<Expression>());
            return Expression.Lambda<Func<T, bool>>(Expression.Or(expression1.Body, invokedExpression), expression1.Parameters);
        }

        public static Expression<Func<T, bool>> And<T>(this Expression<Func<T, bool>> expression1, Expression<Func<T, bool>> expression2)
        {
            var invokedExpression = Expression.Invoke(expression2, expression1.Parameters.Cast<Expression>());
            return Expression.Lambda<Func<T, bool>>(Expression.And(expression1.Body, invokedExpression), expression1.Parameters);
        }

        private static Expression ExpressFilter(Operator opator, Expression left, Expression right)
        {
            switch (opator)
            {
                case Operator.Equal:
                    {
                        return Expression.Equal(left, right);
                    }
                case Operator.GreaterThan:
                    {
                        return Expression.GreaterThan(left, right);
                    }
                case Operator.GreaterThanOrEqual:
                    {
                        return Expression.GreaterThanOrEqual(left, right);
                    }
                case Operator.In:
                    {
                        return ExpressionExpand.In(left, right);
                    }
                case Operator.LessThan:
                    {
                        return Expression.LessThan(left, right);
                    }
                case Operator.LessThanOrEqual:
                    {
                        return Expression.LessThanOrEqual(left, right);
                    }
                case Operator.Like:
                    {
                        return ExpressionExpand.Like(left, right);
                    }
                case Operator.NotEqual:
                    {

                        return Expression.NotEqual(left, right);
                    }
                case Operator.NotIn:
                    {
                        return ExpressionExpand.NotIn(left, right);
                    }
                case Operator.NotLike:
                    {
                        return ExpressionExpand.NotLike(left, right);
                    }
                default: { throw new ArgumentNullException(); }
            }
        }
    }
}